--******************************************************************************
--* Company:     FTDI                                                          *
--* Date:        23 Nov 2004                                                   *
--* Author:      Iain Maciver                                                  *
--*                                                                            *
--* Module Name: freq_cnt_ctl                                                  *
--*                                                                            *
--* Description: Calculates the frequency of an unknown input signal using     *
--*              Morph-IC clk as a reference                                   *
--******************************************************************************

LIBRARY IEEE;
USE IEEE.std_logic_1164.ALL;

ENTITY freq_cnt_ctl IS
  PORT (
        clk50     : in std_logic;  	   -- 50MHz clock input
        FUnknown  : in std_logic;      -- 12Mhz recovered clock input
        rst       : in std_logic;  	   -- Active high reset via AD7
        FT245Din  : in std_logic_vector(7 downto 0); -- Data Bus
        nRXF      : in std_logic;      -- Read Buffer Full
        nTXE      : in std_logic;      -- Write Buffer Empty

        FT245Dout : out std_logic_vector(7 DOWNTO 0); -- Data Bus
        FT245DEn  : out std_logic;     -- Data Bus Enable
        nRD       : out std_logic;     -- Read Strobe
        WR        : out std_logic      -- Write Strobe
        );
END freq_cnt_ctl;

ARCHITECTURE rtl OF freq_cnt_ctl IS

  COMPONENT lpmcnt24
    PORT (clock  : IN STD_LOGIC ;
          cnt_en : IN STD_LOGIC ;
          sclr   : IN STD_LOGIC ;
          aclr   : IN STD_LOGIC ;
          q      : OUT STD_LOGIC_VECTOR (23 DOWNTO 0)
          );
  END COMPONENT;

  COMPONENT lpmcnt26
    PORT (clock  : IN STD_LOGIC ;
          cnt_en : IN STD_LOGIC ;
          sclr   : IN STD_LOGIC ;
          aclr   : IN STD_LOGIC ;
          q      : OUT STD_LOGIC_VECTOR (25 DOWNTO 0)
          );
  END COMPONENT;

  -- State Machine encoding
  type mstate is (IDLE, Rx1, Rx2, Rx3, COUNT, Tx1, Tx2, Tx3, Tx4);

  signal ctl_state : mstate;
  --
  signal high   : std_logic;
  signal low    : std_logic;
  --
  signal TXE    : std_logic;
  signal RXF    : std_logic;
  signal FT_En  : std_logic;
  signal WR_int : std_logic;
  signal RD_int : std_logic;
  --
  signal LdCmdReg       : std_logic;
  signal CmdReg         : std_logic_vector(7 downto 0);
  signal cnt_enable     : std_logic;
  signal window_cnt     : std_logic_vector(25 downto 0);
  signal end_window     : std_logic;
  signal freq_cnt_async : std_logic_vector(23 downto 0);
  signal rst_window     : std_logic;
  signal rst_freq_cnt   : std_logic;
  signal FUnknown_d1    : std_logic;
  signal FUnknown_d2    : std_logic;
  signal posedge_FUnknown : std_logic;
  signal Tx_Status      : std_logic_vector(1 downto 0);
  signal Overflow_async : std_logic;
  signal Freq_Out       : std_logic_vector(23 downto 0);
  --
  signal timeout_enable : std_logic;
  signal timeout_cnt    : std_logic_vector(25 downto 0);
  signal timeout        : std_logic;
  signal remainder_en   : std_logic;
  signal remainder_cnt  : std_logic_vector(23 downto 0);
  --

BEGIN

  high <= '1';
  low  <= '0';

---------------------------------------------------------------------------------
-- Resync incoming signals to 50MHz clk
---------------------------------------------------------------------------------
  r_TXE : process (clk50, rst)
  begin
    if (rst = '1') then
      TXE <= '0';
      RXF <= '0';
    elsif (clk50'event and clk50 = '1') then
      TXE <= not nTXE;
      RXF <= not nRXF;
    end if;
  end process r_TXE;
  
---------------------------------------------------------------------------------
-- Store Comamnd when read from FT Chip
---------------------------------------------------------------------------------
  r_Cmd : process (clk50, rst)
  begin
    if (rst = '1') then
      CmdReg <= "00000000";
    elsif (clk50'event and clk50 = '1') then
      if (LdCmdReg = '1') then
        CmdReg <= FT245Din;
      else
        CmdReg <= CmdReg;
      end if;
    end if;
  end process r_Cmd;
  
  r_FUnknown_del : process (clk50, rst)
  begin
    if (rst = '1') then
      FUnknown_d1 <= '0';
      FUnknown_d2 <= '0';
    elsif (clk50'event and clk50 = '1') then
      FUnknown_d1 <= FUnknown;
      FUnknown_d2 <= FUnknown_d1;
    end if;
  end process r_FUnknown_del;

  -- Detect poitive edge of FUnknown
  posedge_FUnknown <= FUnknown_d1 and not FUnknown_d2;
  
---------------------------------------------------------------------------------
-- Control Counters: Timeout - If not edges are found within the timeout period
--                             this enables the state machine to complete
--                   Window  - Enables the frequency counter for fixed window
--                   FreqCnt - Counts the number of edges in the Window period
---------------------------------------------------------------------------------
  i_timeout : lpmcnt26 PORT MAP(
    clock  => clk50,
    cnt_en => timeout_enable,
    sclr   => rst_window,
    aclr   => rst,
    q      => timeout_cnt
    );
    
  i_window : lpmcnt26 PORT MAP(
    clock  => clk50,
    cnt_en => cnt_enable,
    sclr   => rst_window,
    aclr   => rst,
    q      => window_cnt
    );
    
  i_freq_cnt_async : lpmcnt24 PORT MAP(
    clock  => FUnknown,
    cnt_en => cnt_enable and not Overflow_async,
    sclr   => low,
    aclr   => rst or rst_freq_cnt,
    q      => freq_cnt_async
    );

  p_overflow_async : process (freq_cnt_async)
  begin
    if (freq_cnt_async = "111111111111111111111111") then
      Overflow_async <= '1';
    else
      Overflow_async <= '0';
    end if;
  end process p_overflow_async;
  
  p_end_window : process (window_cnt, timeout_cnt, CmdReg)
  begin
    case CmdReg(2 downto 0) is
      -- 1 sec
      when "100" =>
        if (timeout_cnt = "10111110101111000001111111") then
          timeout <= '1';
        else
          timeout <= '0';
        end if;
        if (window_cnt = "10111110101111000001111111") then
          end_window <= '1';
        else
          end_window <= '0';
        end if;
      -- 500 ms
      when "011" =>
        if (timeout_cnt = "01011111010111100000111111") then
          timeout <= '1';
        else
          timeout <= '0';
        end if;
        if (window_cnt = "01011111010111100000111111") then
          end_window <= '1';
        else
          end_window <= '0';
        end if;
      -- 100 ms
      when "010" =>
        if (timeout_cnt = "00010011000100101100111111") then
          timeout <= '1';
        else
          timeout <= '0';
        end if;
        if (window_cnt = "00010011000100101100111111") then
          end_window <= '1';
        else
          end_window <= '0';
        end if;
      -- 10 ms
      when "001" =>
        if (timeout_cnt = "00000001111010000100011111") then
          timeout <= '1';
        else
          timeout <= '0';
        end if;
        if (window_cnt = "00000001111010000100011111") then
          end_window <= '1';
        else
          end_window <= '0';
        end if;
      -- 1 ms
      when others =>
        if (timeout_cnt = "00000000001100001101001111") then
          timeout <= '1';
        else
          timeout <= '0';
        end if;
        if (window_cnt = "00000000001100001101001111") then
          end_window <= '1';
        else
          end_window <= '0';
        end if;
      --
    end case;
  end process p_end_window;
  
---------------------------------------------------------------------------------
-- State Machine Controller
---------------------------------------------------------------------------------
  sm_freq_cnt : process (clk50, rst)
  begin
    if (rst = '1') then
      ctl_state      <= IDLE;
      cnt_enable     <= '0';
      timeout_enable <= '0';
      remainder_en   <= '0';
      rst_window     <= '0';
      rst_freq_cnt   <= '0';
      FT_En          <= '0';
      RD_int         <= '0';
      WR_int         <= '0';
      LdCmdReg       <= '0';
      Tx_Status      <= "00";
    elsif (clk50'event and clk50 = '1') then
      case (ctl_state) is
        -- Wait until data is available to be received from Rx Buffer
        when IDLE =>
          if (RXF = '1') then
            ctl_state    <= Rx1;
            RD_int       <= '1';
          else
            ctl_state    <= IDLE;
            RD_int       <= '0';
          end if;
          cnt_enable     <= '0';
          timeout_enable <= '0';
          remainder_en   <= '0';
          rst_window     <= '0';
          rst_freq_cnt   <= '0';
          FT_En          <= '0';
          WR_int         <= '0';
          LdCmdReg       <= '0';
          Tx_Status      <= "00";
        -- 
        when Rx1 =>
          ctl_state      <= Rx2;
          cnt_enable     <= '0';
          timeout_enable <= '0';
          remainder_en   <= '0';
          rst_window     <= '0';
          rst_freq_cnt   <= '0';
          FT_En          <= '0';
          RD_int         <= '1';
          WR_int         <= '0';
          LdCmdReg       <= '1';
          Tx_Status      <= "00";
        -- 
        when Rx2 =>
          ctl_state      <= Rx3;
          cnt_enable     <= '0';
          timeout_enable <= '0';
          remainder_en   <= '0';
          rst_window     <= '0';
          rst_freq_cnt   <= '0';
          FT_En          <= '0';
          RD_int         <= '1';
          WR_int         <= '0';
          LdCmdReg       <= '1';
          Tx_Status      <= "00";
        -- If Cmd is X"Bx" start Frequency measurement
        when Rx3 =>
          if (RXF = '1') then
            ctl_state      <= Rx3;
            cnt_enable     <= '0';
            timeout_enable <= '0';
          elsif (CmdReg(7 downto 4) = "1011") then
            ctl_state      <= COUNT;
            cnt_enable     <= posedge_FUnknown;
            timeout_enable <= '1';
          else
            ctl_state      <= IDLE;
            cnt_enable     <= '0';
            timeout_enable <= '0';
          end if;
          rst_window   <= '0';
          remainder_en <= '0';
          rst_freq_cnt <= '0';
          FT_En        <= '0';
          RD_int       <= '0';
          WR_int       <= '0';
          LdCmdReg     <= '0';
          Tx_Status    <= "00";
        -- wait until measurement window has completed
        when COUNT =>
          if (end_window = '1') then
            ctl_state    <= Tx1;
            cnt_enable   <= '0';
            remainder_en <= '0';
            FT_En        <= '1';
          else
            ctl_state    <= COUNT;
            cnt_enable   <= cnt_enable or posedge_FUnknown or timeout;
            remainder_en <= '0';
            FT_En        <= '0';
          end if;
          timeout_enable <= '1';
          rst_window     <= '0';
          rst_freq_cnt   <= '0';
          RD_int         <= '0';
          WR_int         <= '0';
          LdCmdReg       <= '0';
          Tx_Status      <= "00";
        -- Transfer Frequency count to Tx Buffer in four bytes
        when Tx1 =>
          if (TXE = '1') then
            ctl_state    <= Tx2;
            WR_int       <= '1';
          else
            ctl_state    <= Tx1;
            WR_int       <= '0';
          end if;
          cnt_enable     <= '0';
          timeout_enable <= '0';
          remainder_en   <= '0';
          rst_window     <= '0';
          rst_freq_cnt   <= '0';
          FT_En          <= '1';
          RD_int         <= '0';
          LdCmdReg       <= '0';
          Tx_Status      <= Tx_Status;
        -- 
        when Tx2 =>
          ctl_state      <= Tx3;
          cnt_enable     <= '0';
          timeout_enable <= '0';
          remainder_en   <= '0';
          rst_window     <= '0';
          rst_freq_cnt   <= '0';
          FT_En          <= '1';
          RD_int         <= '0';
          WR_int         <= '1';
          LdCmdReg       <= '0';
          Tx_Status      <= Tx_Status;
        -- 
        when Tx3 =>
          ctl_state      <= Tx4;
          cnt_enable     <= '0';
          timeout_enable <= '0';
          remainder_en   <= '0';
          rst_window     <= '0';
          rst_freq_cnt   <= '0';
          FT_En          <= '1';
          RD_int         <= '0';
          WR_int         <= '1';
          LdCmdReg       <= '0';
          Tx_Status      <= Tx_Status;
        -- Return to Idle when last byte has been transferred
        when Tx4 =>
          if (TXE = '1') then
            ctl_state     <= Tx4;
            rst_window    <= '0';
            rst_freq_cnt  <= '0';
            Tx_Status     <= Tx_Status;
          elsif (Tx_Status = "11") then
            ctl_state     <= IDLE;
            rst_window    <= '1';
            rst_freq_cnt  <= '1';
            Tx_Status     <= Tx_Status;
          else
            ctl_state     <= Tx1;
            rst_window    <= '0';
            rst_freq_cnt  <= '0';
            case Tx_Status is
              when "10"   => Tx_Status <= "11";
              when "01"   => Tx_Status <= "10";
              when "00"   => Tx_Status <= "01";
              when others => Tx_Status <= "00";
            end case;
          end if;
          cnt_enable     <= '0';
          timeout_enable <= '0';
          remainder_en   <= '0';
          FT_En          <= '1';
          RD_int         <= '0';
          WR_int         <= '0';
          LdCmdReg       <= '0';
        -- 
        when others =>
          ctl_state      <= IDLE;
          cnt_enable     <= '0';
          timeout_enable <= '0';
          remainder_en   <= '0';
          rst_window     <= '0';
          rst_freq_cnt   <= '0';
          FT_En          <= '0';
          RD_int         <= '0';
          WR_int         <= '0';
          LdCmdReg       <= '0';
          Tx_Status      <= "00";
      end case;
    end if;
  end process sm_freq_cnt;

---------------------------------------------------------------------------------
-- Assign Outputs
---------------------------------------------------------------------------------
  Freq_Out <= freq_cnt_async;

  -- Encode Frequency bytes using upper two bits
  p_FT_Data : process (Tx_Status, Freq_Out)
  begin
    case Tx_Status is
      when "11"   => FT245Dout <= "11" & Freq_Out(23 downto 18);
      when "10"   => FT245Dout <= "10" & Freq_Out(17 downto 12);
      when "01"   => FT245Dout <= "01" & Freq_Out(11 downto 6);
      when others => FT245Dout <= "00" & Freq_Out(5 downto 0);
    end case;
  end process p_FT_Data;

  FT245DEn <= FT_En;
  WR       <= WR_int;
  nRD      <= not RD_int;

END rtl;

